--*****************************************************************************
--* File: lua/modules/ui/game/queue.lua
--* Author: Chris Blackwell / Ted Snook
--* Summary: Construction Queue UI
--*
--* Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
--*****************************************************************************

local UIUtil = import('/lua/modules/ui/uiutil.lua')
local LayoutHelpers = import('/lua/modules/maui/layouthelpers.lua')
local Group = import('/lua/modules/maui/group.lua').Group
local Bitmap = import('/lua/modules/maui/bitmap.lua').Bitmap
local Grid = import('/lua/modules/maui/grid.lua').Grid
local Checkbox = import('/lua/modules/maui/checkbox.lua').Checkbox
local Button = import('/lua/modules/maui/button.lua').Button
local GameCommon = import('/lua/modules/ui/game/gamecommon.lua')
local RadioGroup = import('/lua/modules/maui/mauiutil.lua').RadioGroup
local Tooltip = import('/lua/modules/ui/game/tooltip.lua')

local constructionControl = false
local currentSelection = false
currentQueue = false

# vars added in by mod
local dragging = false
local index = nil			# index of the item in the queue currently being dragged
local originalIndex = false	# original index of selected item (so that UpdateBuildQueue knows where to modify it from)
local oldQueue = false


local currentInfiniteQueueCheckStatus = false
local layoutVar = false
controls = {
    queueGroup = false,
    queueBGCapMin = false,
    queueBGMiddle = false,
    queueBGCapMax = false,
    queueGrid = false,
    scrollQueueFwdButton = false,
    scrollQueueBackButton = false,
    infiniteQueueToggle = false,
    pauseToggle = false,
    mfdControl = false,
    ordersControl = false,
    controlClusterGroup = false,
}

local function CreateQueueGrid(parent)
    controls.queueGrid = Grid(parent, 62, 62)
    controls.queueGrid:SetName("Queue Grid")
    controls.queueGrid:AppendRows(1)
end

function ToggleUnitPause()
    controls.pauseToggle:ToggleCheck()
end

function CreateControls()
    if not controls.queueGroup then
        controls.queueGroup = Group(controls.controlClusterGroup, "queueGroup")
    end
    if not controls.queueBGCapMin then
        controls.queueBGCapMin = Bitmap(controls.queueGroup)
    end
    if not controls.queueBGCapMax then
        controls.queueBGCapMax = Bitmap(controls.queueGroup)
    end
    if not controls.queueBGMiddle then
        controls.queueBGMiddle = Bitmap(controls.queueGroup)
    end
    if not controls.queueGrid then
        CreateQueueGrid(controls.queueBGMiddle)
    end

    if not controls.queueScrollBG then
        controls.queueScrollBG = Bitmap(controls.queueGroup)
    end
    if controls.scrollQueueBackButton then controls.scrollQueueBackButton:Destroy() end
    controls.scrollQueueBackButton = UIUtil.CreateButton(controls.queueBGCapMin
        , '/game/small-side_scr/arrow-left_scr_up.dds'
        , '/game/small-side_scr/arrow-left_scr_down.dds'
        , '/game/small-side_scr/arrow-left_scr_over.dds'
        , '/game/small-side_scr/arrow-left_scr_dis.dds'
        , nil, nil, nil, nil, "UI_Arrow_Click")
    if controls.scrollQueueFwdButton then controls.scrollQueueFwdButton:Destroy() end
    controls.scrollQueueFwdButton = UIUtil.CreateButton(controls.queueBGCapMin
        , '/game/small-side_scr/arrow-right_scr_up.dds'
        , '/game/small-side_scr/arrow-right_scr_down.dds'
        , '/game/small-side_scr/arrow-right_scr_over.dds'
        , '/game/small-side_scr/arrow-right_scr_dis.dds'
        , nil, nil, nil, nil, "UI_Arrow_Click")
    if controls.infiniteQueueToggle then controls.infiniteQueueToggle:Destroy() end
    controls.infiniteQueueToggle = UIUtil.CreateCheckbox(controls.queueBGCapMin,
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_up.dds'),
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_down.dds'),
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_over.dds'),
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_down.dds'),
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_dis.dds'),
        UIUtil.SkinnableFile('/game/infinite_btn/infinite_btn_dis.dds'),
        "UI_Construct_Pause")
    if controls.pauseToggle then controls.pauseToggle:Destroy() end
    controls.pauseToggle = UIUtil.CreateCheckbox(controls.queueBGCapMin,
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_up.dds'),
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_down.dds'),
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_over.dds'),
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_down.dds'),
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_dis.dds'),
        UIUtil.SkinnableFile('/game/pause_btn/pause_btn_dis.dds'),
        "UI_Construct_Infinite")
end

function CommonLogic()
    controls.scrollQueueBackButton.OnClick = function(self)
        controls.queueGrid:ScrollLines("Horz", -1)
    end
    controls.scrollQueueBackButton:Disable()
    controls.scrollQueueFwdButton.OnClick = function(self)
        controls.queueGrid:ScrollLines("Horz", 1)
    end
    controls.scrollQueueFwdButton:Disable()
    controls.infiniteQueueToggle.OnClick = function(self)
        if currentInfiniteQueueCheckStatus then
            currentInfiniteQueueCheckStatus = false
        else
            currentInfiniteQueueCheckStatus = true
        end
        OnInfiniteQueueToggle(currentInfiniteQueueCheckStatus)
    end
    controls.infiniteQueueToggle.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "infinite_toggle")
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
        Checkbox.HandleEvent(self, event)
    end
    controls.pauseToggle.OnCheck = function(self, checked)
        SetPaused(currentSelection, checked)
    end
    controls.pauseToggle.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "pause")
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
        Checkbox.HandleEvent(self, event)
    end
end

function SetLayout(layout)
    CreateControls()
    layoutVar = layout
    import(UIUtil.GetLayoutFilename('queue')).SetLayout()
    CommonLogic()
end

function SetupQueueControl(parent, inMFDControl, inOrdersControl)
    controls.mfdControl = inMFDControl
    controls.ordersControl = inOrdersControl
    controls.controlClusterGroup = parent

    SetLayout(UIUtil.currentLayout)
    return controls.queueGroup
end

function SetQueueState(state)
    controls.queueGroup.Mini(state)
end

function SetInfiniteQueueToggle(state)
    if state then
        controls.infiniteQueueToggle:Enable()
    else
        controls.infiniteQueueToggle:Disable()
    end
end

function SetInfinityCheck(state)
    currentInfiniteQueueCheckStatus = state
    controls.infiniteQueueToggle:SetCheck(state)
end

-- command queue is a table of unit id's and how many to build, in sequence to be built
local currentCommandQueue = false

function CreateQueueIcon(parent)
    local bg = Bitmap(parent) -- TODO unit specific BG here
    bg:SetSolidColor('00000000')
    #bg.Width:Set(GameCommon.iconBmpWidth)
    #bg.Height:Set(GameCommon.iconBmpHeight)
    local icon = Bitmap(bg)
    LayoutHelpers.AtLeftTopIn(icon, bg)
    #LayoutHelpers.FillParent(icon, bg)
    local text = UIUtil.CreateText(icon, "", 18, UIUtil.bodyFont)
    LayoutHelpers.AtLeftTopIn(text, icon, 5, 5)
    if layoutVar == 'minimal' then
        icon.Height:Set(48)
        icon.Width:Set(48)
    end
    text:SetDropShadow(true)
    text:SetColor('white')
    text:DisableHitTest()
    bg.text = text
    bg.unitIcon = icon
    bg:Hide()

    return bg
end

function ClearQueueGrid()
    ClearSessionExtraSelectList()
    controls.queueGrid:DeleteAndDestroyAll(true)
    controls.queueGrid:AppendRows(1, true)
    controls.queueGrid:EndBatch()
end

function PopulateQueueGrid(numIcons)
    if numIcons == 0 then
        controls.scrollQueueFwdButton:Disable()
        controls.scrollQueueBackButton:Disable()
        ClearQueueGrid()
        return
    end
    local cols, rows = controls.queueGrid:GetDimensions()
    if numIcons > cols then
        controls.queueGrid:AppendCols(numIcons - cols, true)
        for i = cols + 1, numIcons do
            controls.queueGrid:SetItem(CreateQueueIcon(controls.queueGrid), i, 1, true)
        end
    elseif numIcons < cols then
        for i = cols, numIcons + 1, -1 do
            controls.queueGrid:DestroyItem(i, 1, true)
            controls.queueGrid:DeleteCol(i, true)
        end
    end
end

function MoveItemInQueue(queue, indexfrom, indexto)
	if indexfrom <= indexto then
		#take indexfrom out and shunt all indices from indexfrom to indexto up one
		moveditem = queue[indexfrom]
		for i = indexfrom, (indexto - 1) do
			queue[i] = queue[i+1]
		end
		queue[indexto] = moveditem
	else
		#take indexfrom out and shunt all indices from indexto to indexfrom down one
		moveditem = queue[indexfrom]
		for i = 0, (indexfrom - indexto - 1) do
			queue[indexfrom - i] = queue[indexfrom-i-1]
		end
		queue[indexto] = moveditem
	end
	return queue
end


function UpdateBuildList(newqueue, from)
	LOG("FROM = ", from)
      # The way this does this is UGLY but I can only find functions to remove things from the build queue and to add them at the end
	# Thus the only way I can see to modify the build queue is to delete it back to the point it is modified from (the from argument) and then 
      # add the modified version back in. Unfortunately this causes a momentary 'skip' in the displayed build cue as it is deleted and replaced
	for i = from, table.getn(oldQueue)  do
		# I can only get for loop to count forwards so do some ugly math to get it counting backwards
		reverse_index = table.getn(oldQueue) - i + from
		DecreaseBuildCountInQueue(reverse_index, oldQueue[reverse_index].count)	
	end
	for i = from, table.getn(newqueue)  do
		LOG("ID: ", newqueue[i].id)
            blueprint = __blueprints[newqueue[i].id]
            if blueprint.General.UpgradesFrom == 'none' then
		    IssueBlueprintCommand("UNITCOMMAND_BuildFactory", newqueue[i].id, newqueue[i].count)
		else
		    IssueBlueprintCommand("UNITCOMMAND_Upgrade", newqueue[i].id, 1, false)
            end
	end

	# FIX ME: the factory will pause (I am guessing due to some time out) before it starts building again, once the build queue has been modified
end

function ButtonReleaseCallback(event)
    if dragging == true then
         # mouse button released so end drag
         dragging = false

	   if originalIndex <= index then
	        first_modified_index = originalIndex
	   else 
		  first_modified_index = index
	   end

	   # on the release of the mouse button we want to update the ACTUAL build queue that the factory does. So far, only the UI has been changed,
	   # the original build queue will be reverted to once the factory starts building the next unit unless we change it.
	   UpdateBuildList(modifiedQueue, first_modified_index)

	   # nothing is now selected
	   index = nil    
    end  

end

function SetQueueGrid(queue, selection)
    currentQueue = queue
    
    local numIcons = table.getn(queue)

    PopulateQueueGrid(numIcons)

    for i = 1, numIcons do
        local icon = controls.queueGrid:GetItem(i, 1)
        if not icon then
            LOG("SetQueueGrid attempt to get item ", i, " but it's not there.")
            continue
        end
        local bpid = queue[i].id
        local blueprint = __blueprints[bpid]
        local iconName, upIconName, downIconName, overIconName = GameCommon.GetCachedUnitIconFileNames(blueprint)
        icon.unitIcon:SetTexture(upIconName)
        icon.text:SetText(tostring(queue[i].count))
        icon._blueprintId = bpid
        icon._index = i -- store array index for determining who was clicked on

        icon.HandleEvent = function(self, event)
            if event.Type == 'MouseEnter' then
                local blueprint = __blueprints[self._blueprintId]
                import('/lua/modules/ui/game/unitviewDetail.lua').Show(blueprint, currentSelection[1], self._blueprintId)

		    # if we have entered the button and are dragging something then we want to replace it with what we are dragging
		    if dragging == true then
                    # move item from old location (index) to new location (this button's index)
		        modifiedQueue = MoveItemInQueue(currentQueue, index, self._index) 
                    # since the currently selected button has now moved, update the index
                    index = self._index
		        # update buttons in the UI
                    SetQueueGrid(modifiedQueue, selection)
                end
            elseif event.Type == 'MouseExit' then
                    import('/lua/modules/ui/game/unitviewDetail.lua').Hide()
            elseif event.Type == 'ButtonPress' or event.Type == 'ButtonDClick' then
                local count = 1
                if event.Modifiers.Ctrl == true and event.Modifiers.Shift == nil then
                    count = 5
                elseif event.Modifiers.Shift == true and event.Modifiers.Ctrl == nil then
                    count = 5
                elseif event.Modifiers.Shift == true and event.Modifiers.Ctrl == true then
                    count = 5
                end

                if event.Modifiers.Left then
			  gameParent = import('gamemain.lua').GetGameParent()
			  gameParent.HandleEvent = function(self, event)
        	        if event.Type == 'ButtonRelease' then
				import('/lua/modules/ui/game/queue.lua').ButtonReleaseCallback(event)
        		  end 

    		 end 
			  # only perform drag and drop on factory build queues.
                    if EntityCategoryContains(categories.FACTORY, currentSelection[1]) then
				# only perform drag and drop on factory build queues.
                        # left button pressed so start dragging procedure
                        dragging = true
                        index = self._index
			      originalIndex = index

                        # copy un modified queue so that current build order is recorded (for deleting it)
                        oldQueue = table.copy(currentQueue)
			  else
				IncreaseBuildCountInQueue(self._index, count)
			  end
                else
                    DecreaseBuildCountInQueue(self._index, count)
                end
            end

		if event.Type == 'ButtonRelease' then
		end 
            Bitmap.HandleEvent(self, event)
        end

    end
    controls.queueGrid:ScrollSetTop('Horz', 1)
    controls.queueGrid:EndBatch()
    local visCols, visRows = controls.queueGrid:GetVisible()
    if (visCols * visRows) >= numIcons then
        controls.scrollQueueFwdButton:Disable()
        controls.scrollQueueBackButton:Disable()
    else
        controls.scrollQueueFwdButton:Enable()
        controls.scrollQueueBackButton:Enable()
    end
end

function RemovedLastQueuedUnit(bpId)
    if bpId and currentQueue then
        local i = table.getn(currentQueue)
        while i > 0 do
            local icon = controls.queueGrid:GetItem(i, 1)
            if icon._blueprintId == bpId then
                DecreaseBuildCountInQueue(icon._index, 1)
                break
            end
            i = i - 1
        end
    end
end

function SetSelectionIcons(sortedUnitTable, sortedObjectTables)
    ClearQueueGrid()

    local numTypes = table.getn(sortedUnitTable)
    local numIcons = 0

    if numTypes then
        local iconBPID = nil
        local cols, rows = controls.queueGrid:GetDimensions()
        numIcons = table.getn(sortedUnitTable)

        PopulateQueueGrid(numIcons)

        local i = 1
        for index, v in sortedUnitTable do
            local icon = controls.queueGrid:GetItem(i, 1)
            if not icon then
                LOG("SetQueueGrid attempt to get item ", i, " but it's not there.")
            end
            local blueprint = __blueprints[index]
            local iconName, upIconName, downIconName, overIconName = GameCommon.GetCachedUnitIconFileNames(blueprint)
            icon.unitIcon:SetTexture(upIconName)
            icon.text:SetText(v)
            icon._unitType = index
            i = i + 1

            icon.HandleEvent = function(self, event)
                if event.Type == 'ButtonPress' or event.Type == 'ButtonDClick' then
                    if event.Modifiers.Left then
                        SelectUnits(sortedObjectTables[self._unitType])
                    else
                        local index = 1
                        local newUnitSelectTable = {}
                        for i, v in sortedObjectTables do
                            if i != self._unitType then
                                for y, z in v do
                                    newUnitSelectTable[index] = z
                                    index = index + 1
                                end
                            end
                        end
                        SelectUnits(newUnitSelectTable)
                    end
                end
                Bitmap.HandleEvent(self, event)
            end
        end
    end


    controls.queueGrid:EndBatch()
    local visCols, visRows = controls.queueGrid:GetVisible()
    if (visCols * visRows) >= numIcons then
        controls.scrollQueueFwdButton:Disable()
        controls.scrollQueueBackButton:Disable()
    else
        controls.scrollQueueFwdButton:Enable()
        controls.scrollQueueBackButton:Enable()
    end
end

function EnablePauseToggle()
    controls.pauseToggle:Enable()
end

-- called on new selection, determines what, if anything, goes in the order queue
function CheckForOrderQueue(selection, newSelection)
    currentSelection = newSelection
    if controls.pauseToggle then controls.pauseToggle:Disable() end
    controls.pauseToggle:SetCheck(GetIsPaused(currentSelection),true)
    if currentSelection[1] and EntityCategoryContains(categories.FACTORY, currentSelection[1]) then
        controls.infiniteQueueToggle:Enable()
    else
        controls.infiniteQueueToggle:Disable()
    end
    -- only show a queue if single item is selected
    if table.getn(selection) == 1 then
        currentCommandQueue = SetCurrentFactoryForQueueDisplay(selection[1])
        -- render the command queue
        if currentCommandQueue then
            SetQueueGrid(currentCommandQueue, selection)
        else
            ClearQueueGrid()
        end
        SetQueueState(false)
    else
        ClearCurrentFactoryForQueueDisplay()
        ClearQueueGrid()
    end
end

function OnQueueChanged(newQueue)
    if newQueue then
        SetQueueGrid(newQueue)
    else
        ClearQueueGrid()
    end
end

function OnInfiniteQueueToggle(checked)
    controls.infiniteQueueToggle:SetCheck(checked)
    for i,v in currentSelection do
        if checked then
            v:ProcessInfo('SetRepeatQueue', 'true')
        else
            v:ProcessInfo('SetRepeatQueue', 'false')
        end
    end
end

function ToggleEnhancementWindow(state, enhancementWindow)
    if state then
        if UIUtil.currentLayout == 'bottom' then
            controls.queueBGCapMin.Left:Set(function() return enhancementWindow.Right() - 23 end)
        end
    else
        if UIUtil.currentLayout == 'bottom' then
            LayoutHelpers.AtLeftIn(controls.queueBGCapMin, controls.constructionGroup, -2)
            LayoutHelpers.AtLeftIn(controls.queueBGCapMin, controls.queueGroup, -2)
        end
    end
end
